﻿using GalaSoft.MvvmLight;
using SciChart.Charting.Model.DataSeries;
using SciChart.Charting.Visuals;
using SciChart.Charting.Visuals.Axes;
using SciChart.Charting.Visuals.RenderableSeries;
using SciChart.Data.Model;
using SciChart.Data.Numerics;
using SciChartPerfTester.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Data;
using System.Windows.Media;
using System.Windows.Threading;

namespace SciChartPerfTester.ViewModel
{
	public class ChartViewModel : ViewModelBase
	{
		private const int _MaxStrokeCount = 7;

		private static readonly Color[] _StrokeColors = new[]
		{
			Colors.Green, Colors.Blue, Colors.Indigo, Colors.Violet, Colors.Red,
			Colors.Orange, Colors.Yellow
		};

		// different strategies for updating - TODO: should be enum for various strategies such as single or update vs append point vs append list
		private bool _UseUpdate = true;
		private bool _UseLists = false;



		// for MinMax DataType -- not really useful now that there are multiple types
		private const double _MinSourceValue = -0.50;
		private const double _MaxSourceValue = 1.50;
		private const double _SourceDelta = 0.005;

		private object _syncObject = new object();
		private Task<bool> _updateTask = null;

		private int _gapSize = 50;
		private double _sourceValue = 0;
		private double _sourceDelta = _SourceDelta;
		private long _currentIndex;
		private int _renderSeriesIndex = 0;

		// pts per second
		private double _dataFrequency;

		static ChartViewModel()
		{
			InitializeSamples();
		}

		public ChartViewModel(int numPointsPerSeries)
		{
			if (IsInDesignMode)
			{
				// Code runs in Blend --> create design time data.
				MaxFrameRate = 30.0;
			}
			else
			{
				// Code runs "for real"
			}

			RenderableSeries = new ObservableCollection<IRenderableSeries>();
			NumPointsPerSeries = numPointsPerSeries;
			UpdateDataGenerationParameters();
		}

		private bool _isRunning;
		public bool IsRunning { get { return _isRunning; } private set { Set(nameof(IsRunning), ref _isRunning, value); } }

		private DataFunction _dataFunction = DataFunction.SineWave;
		public DataFunction DataFunction
		{
			get { return _dataFunction; }
			set
			{
				if (Set(nameof(DataFunction), ref _dataFunction, value))
				{
					ReinitializeSeries();
					UpdateDataGenerationParameters();
				}
			}
		}

		private int _numPointsPerSeries;
		public int NumPointsPerSeries
		{
			get
			{
				return _numPointsPerSeries;
			}
			set
			{
				if (Set(nameof(NumPointsPerSeries), ref _numPointsPerSeries, value))
				{
					ReinitializeSeries();
					RaisePropertyChanged(nameof(XRange));
				}
			}
		}

		private int _numSeries;
		public int NumSeries
		{
			get { return _numSeries; }
			set { if (Set(nameof(NumSeries), ref _numSeries, value)) { ReinitializeSeries(); } }
		}

		public ObservableCollection<IRenderableSeries> RenderableSeries { get; private set; }

		private List<XyDataSeries<long, double>> DataSeriesList { get; set; }

		public IRange XRange { get { return new DoubleRange(0, NumPointsPerSeries); } }

		private IRange _yRange;
		public IRange YRange { get { return _yRange; } set { Set(nameof(YRange), ref _yRange, value); } }

		private int _chartTimeInMs;
		public int ChartTimeInMs
		{
			get { return _chartTimeInMs; }
			set
			{
				if (Set(nameof(ChartTimeInMs), ref _chartTimeInMs, value))
				{
					ReinitializeSeries();
					RaisePropertyChanged(nameof(XRange));
				}
			}
		}

		private int _timeToSleepInMs;
		public int TimeToSleepInMs
		{
			get { return _timeToSleepInMs; }
			set { Set(nameof(TimeToSleepInMs), ref _timeToSleepInMs, value); }
		}

		private Visibility _diagnosticsVisibility;
		public Visibility DiagnosticsVisibility { get { return _diagnosticsVisibility; } set { Set(nameof(DiagnosticsVisibility), ref _diagnosticsVisibility, value); } }

		private bool _antiAliasing;
		public bool AntiAliasing
		{
			get { return _antiAliasing; }
			set { if (Set(nameof(AntiAliasing), ref _antiAliasing, value)) { ReinitializeSeries(); } }
		}

		private int _strokeThickness;
		public int StrokeThickness
		{
			get { return _strokeThickness; }
			set { if (Set(nameof(StrokeThickness), ref _strokeThickness, value)) { ReinitializeSeries(); } }
		}

		private double _maxFrameRate = 30.0;
		public double MaxFrameRate
		{
			get { return _maxFrameRate; }
			set { if (Set(nameof(MaxFrameRate), ref _maxFrameRate, value)) { ReinitializeSeries(); } }
		}


		public static int MaxStrokeCount { get { return _MaxStrokeCount; } }

		private ResamplingMode _resamplingMode = ResamplingMode.None;
		public ResamplingMode ResamplingMode
		{
			get { return _resamplingMode; }
			set { if (Set(nameof(ResamplingMode), ref _resamplingMode, value)) { ReinitializeSeries(); } }
		}

		private RenderPriority _renderPriority;
		public RenderPriority RenderPriority
		{
			get { return _renderPriority; }
			set
			{
				if (Set(nameof(RenderPriority), ref _renderPriority, value))
				{
					if (SciChartSurface != null)
					{
						SciChartSurface.RenderPriority = value;
						ReinitializeSeries();
					}
				}
			}
		}



		private string _fpsString;
		public string FpsString { get { return _fpsString; } set { Set(nameof(FpsString), ref _fpsString, value); } }

		public void Start()
		{
			lock (_syncObject)
			{
				Log.D("Starting chart...");
				if (IsRunning)
				{
					Log.D("Attempting to start an already started VM.");
					return;
				}

				IsRunning = true;

				_lastRenderTime = DateTime.Now;

				_updateTask = Task.Factory.StartNew(() => UpdateTaskFunction(), TaskCreationOptions.LongRunning);
			}
		}

		public void Stop()
		{
			lock (_syncObject)
			{
				Log.D("Stopping chart...");
				if (!IsRunning)
				{
					Log.D("Attempting to stop a non-running VM.");
					return;
				}

				IsRunning = false;

				// wait at least as long as the time it should sleep before trying to dispose the Task
				if (!_updateTask.Wait(TimeSpan.FromMilliseconds(TimeToSleepInMs * 1.5)))
				{
					try
					{
						_updateTask.Dispose();
					}
					catch (Exception ex)
					{
						Log.D($"Exception on stopping: {ex}");
					}
				}
			}
		}

		private void ReinitializeSeries()
		{
			// if _updateTask is null, we are not controlling ourselves
			bool restartRequired = IsRunning && _updateTask != null;
			if (restartRequired)
				Stop();

			// 5% of total number of points
			_gapSize = (int)Math.Round(NumPointsPerSeries * 0.05);

			_currentIndex = 0;
			_dataFrequency = NumPointsPerSeries / (ChartTimeInMs * 0.001);

			DataSeriesList = new List<XyDataSeries<long, double>>(NumSeries);

			var xVals = Enumerable.Range(0, NumPointsPerSeries + 1).Select(i => (long)i);
			var yVals = xVals.Select(i => double.NaN);

			RenderableSeries.Clear();
			for (int i = 0; i < NumSeries; i++)
			{
				var dataSeries = new XyDataSeries<long, double>() { FifoCapacity = NumPointsPerSeries };
				dataSeries.Append(xVals, yVals);
				DataSeriesList.Add(dataSeries);

				FastLineRenderableSeries renderableSeries = new FastLineRenderableSeries()
				{
					DataSeries = dataSeries,
					AntiAliasing = AntiAliasing,
					ResamplingMode = ResamplingMode,
					StrokeThickness = StrokeThickness,
					Stroke = _StrokeColors[i] // i should not be > _MaxStrokeCount
				};

				RenderableSeries.Add(renderableSeries);
			}

			if (restartRequired)
				Start();
		}


		private bool UpdateTaskFunction()
		{
			try
			{
				while (IsRunning)
				{
					Tick(DateTime.Now);
					Thread.Sleep(TimeToSleepInMs);
				}
			}
			catch (Exception ex)
			{
				Log.D("Exception caught: " + ex);
			}

			Log.D("Update function complete.");
			return true;
		}

		// used to determine exactly how many points we need to render on a given update
		// based on our desired frequency and how much time actually passed since last render
		private DateTime _lastRenderTime = DateTime.Now;
		private double _partialPoints = 0.0;

		private object _syncRoot = new object();

		public void Tick(DateTime now)
		{
			lock (_syncRoot)
			{
				var timeToRenderFor = now.Subtract(_lastRenderTime);
				var numberOfPointsToRender = _dataFrequency * timeToRenderFor.TotalSeconds;

				// keep track of the decimal portion of number of points to render
				_partialPoints += numberOfPointsToRender - (int)numberOfPointsToRender;
				int numberOfPointsToRenderAsInt = (int)numberOfPointsToRender;

				// when we have a whole point from decimal portions, add that to the number to render
				if (_partialPoints >= 1.0)
				{
					// add the integer portion of partial points
					numberOfPointsToRenderAsInt += (int)_partialPoints;

					// remove int portion and keep remaining decimals
					_partialPoints -= (int)_partialPoints;
				}

				AppendPointsWithSuspendUpdates(numberOfPointsToRenderAsInt);

				// remember the last time we rendered
				_lastRenderTime = now;
			}
		}

		private SciChartSurface _sciChartSurface;
		public SciChartSurface SciChartSurface
		{
			get { return _sciChartSurface; }
			set
			{
				if (Set(nameof(SciChartSurface), ref _sciChartSurface, value))
				{
					SciChartSurface.RenderPriority = RenderPriority;
				}
			}
		}



		private void AppendPointsWithSuspendUpdates(int numPointsToRender)
		{
			// can happen on initialization
			if (SciChartSurface == null)
			{
				Log.D("SciChartSurface is null!");
			}
			else
			{
				using (var surfaceSuspender = SciChartSurface.SuspendUpdates())
				{
					AppendPoints(numPointsToRender);
				}
			}
		}

		private void AppendPoints(int numPointsToRender)
		{
			if (_UseLists)
			{
				AppendPointsViaAppendList(numPointsToRender);
			}
			else if (_UseUpdate)
			{
				AppendPointsViaUpdate(numPointsToRender);
			}
			else
			{
				AppendPointsViaAppendPoint(numPointsToRender);
			}
		}

		private void AppendPointsViaUpdate(int numPointsToRender)
		{
			for (int j = 0; j < numPointsToRender; j++)
			{
				if (_currentIndex >= NumPointsPerSeries - 1)
				{
					_currentIndex = 0;
					_renderSeriesIndex = (_renderSeriesIndex + 1) % NumSeries;
				}

				double voltage = UpdateVoltage();

				try
				{
					DataSeriesList[_renderSeriesIndex].Update(_currentIndex, voltage);
					DataSeriesList[_renderSeriesIndex].Update((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					if (NumSeries > 1)
					{
						int prevIndex = (_renderSeriesIndex - 1) % NumSeries;
						if (prevIndex < 0) prevIndex += NumSeries;
						// need to clear previously visible data
						DataSeriesList[prevIndex].Update((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					}
				}
				catch (Exception ex)
				{
					Log.D("Exception caught: " + ex);
					Log.D("_currentIndex: " + _currentIndex);
					_currentIndex = 0;
					_renderSeriesIndex = 0;
				}
				_currentIndex++;
			}
		}

		private void AppendPointsViaAppendPoint(int numPointsToRender)
		{
			for (int j = 0; j < numPointsToRender; j++)
			{

				if (_currentIndex >= NumPointsPerSeries - 1)
				{
					_currentIndex = 0;
					_renderSeriesIndex = (_renderSeriesIndex + 1) % NumSeries;
				}

				// Get the next voltage and time, and append to the chart
				double voltage = UpdateVoltage();
				try
				{
					DataSeriesList[_renderSeriesIndex].Append(_currentIndex, voltage);
					DataSeriesList[_renderSeriesIndex].Append((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					if (NumSeries > 1)
					{
						// need to clear previously visible data
						int prevIndex = (_renderSeriesIndex - 1) % NumSeries;
						if (prevIndex < 0) prevIndex += NumSeries;
						DataSeriesList[prevIndex].Append((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					}
				}
				catch (Exception ex)
				{
					Log.D("Exception caught: " + ex);
					Log.D("_currentIndex: " + _currentIndex);
					_currentIndex = 0;
					_renderSeriesIndex = 0;
				}
				_currentIndex++;
			}
		}



		private void AppendPointsViaAppendList(int numPointsToRender)
		{
			throw new NotImplementedException("Haven't implemented this with dynamic series count");

			var xVals = new List<long>();
			var yValsCurr = new List<double>();
			var yValsPrev = new List<double>();


			for (int j = 0; j < numPointsToRender; j++)
			{

				if (_currentIndex >= NumPointsPerSeries - 1)
				{
					_currentIndex = 0;
					_renderSeriesIndex = (_renderSeriesIndex + 1) % NumSeries;
				}

				double voltage = UpdateVoltage();

				try
				{
					xVals.Add(_currentIndex);
					yValsCurr.Add(voltage);
					yValsPrev.Add(double.NaN);
				}
				catch (Exception ex)
				{
					Log.D("Exception caught: " + ex);
					Log.D("_currentIndex: " + _currentIndex);
					_currentIndex = 0;
					_renderSeriesIndex = 0;
				}
				_currentIndex++;
			}

			// TODO: Not implemented

		}

		private double UpdateVoltage()
		{
			switch (DataFunction)
			{
				case DataFunction.CosineWave:
					{
						_sourceValue += _sourceDelta;
						return Math.Cos(_sourceValue);
					}
				case DataFunction.MinMax:
					{
						double voltage = _sourceValue += _sourceDelta;
						if (_sourceValue >= _MaxSourceValue)
						{
							_sourceDelta = -1 * _SourceDelta;
						}
						else if (_sourceValue <= _MinSourceValue)
						{
							_sourceDelta = _SourceDelta;
						}
						return voltage;
					}
				case DataFunction.SineWave:
					{
						_sourceValue += _sourceDelta;
						return Math.Sin(_sourceValue);
					}
				case DataFunction.SimulatedEcg:
					{
						return GeneratePointValueFor(_currentIndex);
					}
				default:
					return 0.0;
			}
		}

		private void UpdateDataGenerationParameters()
		{
			switch (DataFunction)
			{
				case DataFunction.CosineWave:
					{
						_sourceDelta = 0.01;
						_sourceValue = 0.0;
						YRange = new DoubleRange(-1.1, 1.1);
						break;
					}
				case DataFunction.MinMax:
					{
						_sourceDelta = 0.01;
						_sourceValue = 0.0;
						YRange = new DoubleRange(_MinSourceValue, _MaxSourceValue);
						break;
					}
				case DataFunction.SineWave:
					{
						_sourceDelta = 0.01;
						_sourceValue = 0.0;
						YRange = new DoubleRange(-1.1, 1.1);
						break;
					}
				case DataFunction.SimulatedEcg:
					{
						_sourceValue = 0.0;
						YRange = new DoubleRange(_MinSourceValue, _MaxSourceValue);
						break;
					}
				default:
					{
						_sourceDelta = 0.0;
						_sourceValue = 0.0;
						YRange = new DoubleRange(_MinSourceValue, _MaxSourceValue);
						break;

					}
			}
		}

		public static double GeneratePointValueFor(long index)
		{
			int i = (int)(index % _EcgSample1.Count);
			return _EcgSample1.ElementAt(i);
		}

		private static List<double> _EcgSample1;
		private static void InitializeSamples()
		{
			Assembly assembly = Assembly.GetExecutingAssembly();
			_EcgSample1 = new List<double>();
			foreach (var sample in Properties.Resources.ECG_Sample1.Split(new string[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries))
			{
				try
				{
					_EcgSample1.Add(Double.Parse(sample));
				}
				catch
				{
					// suppress errors
				}
			}
		}
	}
}