using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Command;
using SciChart.Charting.Visuals;
using SciChart.Data.Numerics;
using SciChart.Drawing.DirectX.Context.D3D10;
using SciChartPerfTester.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Windows.Threading;

namespace SciChartPerfTester.ViewModel
{
	/// <summary>
	/// This class contains properties that the main View can data bind to.
	/// <para>
	/// Use the <strong>mvvminpc</strong> snippet to add bindable properties to this ViewModel.
	/// </para>
	/// <para>
	/// You can also use Blend to data bind with the tool's support.
	/// </para>
	/// <para>
	/// See http://www.galasoft.ch/mvvm
	/// </para>
	/// </summary>
	public class MainViewModel : ViewModelBase
	{
		private Task<bool> _updateTask = null;

		public MainViewModel()
		{
			if (IsInDesignMode)
			{
				// Code runs in Blend --> create design time data.
			}
			else
			{
				//SciChart.Core.Utility.SciChartDebugLogger.Instance.SetLogger(new Log());

				Log.D($"Direct3D10CompatibilityHelper.HasDirectX10CapableGpu: {Direct3D10CompatibilityHelper.HasDirectX10CapableGpu}");
				Log.D($"Direct3D10CompatibilityHelper.HasDirectX10RuntimeInstalled: {Direct3D10CompatibilityHelper.HasDirectX10RuntimeInstalled}");
				Log.D($"Direct3D10CompatibilityHelper.IsSupportedOperatingSystem: {Direct3D10CompatibilityHelper.IsSupportedOperatingSystem}");
				Log.D($"Direct3D10CompatibilityHelper.SupportsDirectX10: {Direct3D10CompatibilityHelper.SupportsDirectX10}");
			}

			StartChartsCommand = new RelayCommand(DoStartCharts);
			StopChartsCommand = new RelayCommand(DoStopCharts);

			Charts = new ObservableCollection<ChartViewModel>();
			InitializeDefaults();
			InitializeCharts();
		}

		private void InitializeDefaults()
		{
			NumCharts = 8;
			TimeToSleepInMs = 20;


			// 200 points per second
			ChartTimeInMs = 7000; // 5 seconds
			NumPointsPerChart = 2000;
			NumSeriesPerChart = 2;

			ShowDiagnostics = true;
			UseSingleTaskForUpdate = false;

			AntiAliasing = true;
			StrokeThickness = 2;
			RenderPriority = RenderPriority.Low;
			ResamplingMode = ResamplingMode.Auto;
			MaxFrameRate = 30.0;

			DataFunction = DataFunction.SimulatedEcg;
		}

		public static int MaxStrokeCount { get { return ChartViewModel.MaxStrokeCount; } }

		public int NumRows
		{
			get
			{
				return NumCharts <= 8 ? NumCharts
					: NumCharts <= 16 ? (int)Math.Ceiling(NumCharts * 0.5)
					: NumCharts <= 24 ? (int)Math.Ceiling(NumCharts * 0.33333333)
					: NumCharts <= 32 ? (int)Math.Ceiling(NumCharts * 0.25)
					: NumCharts <= 40 ? (int)Math.Ceiling(NumCharts * 0.2)
					: (int)Math.Ceiling(NumCharts * 0.1666666);
			}
		}
		public int NumColumns
		{
			get
			{
				return NumCharts <= 8 ? 1
					: NumCharts <= 16 ? 2
					: NumCharts <= 24 ? 3
					: NumCharts <= 32 ? 4
					: NumCharts <= 40 ? 5
					: 6;
			}
		}


		public ICommand StartChartsCommand { get; private set; }
		public ICommand StopChartsCommand { get; private set; }

		private bool _isRunning;
		public bool IsRunning
		{
			get { return _isRunning; }
			set { if (Set(nameof(IsRunning), ref _isRunning, value)) { RaisePropertyChanged(nameof(IsNotRunning)); } }
		}

		public bool IsNotRunning { get { return !_isRunning; } }

		private int _numCharts;
		public int NumCharts
		{
			get { return _numCharts; }
			set { if (Set(nameof(NumCharts), ref _numCharts, value)) { InitializeCharts(); } }
		}

		private int _numSeries;
		public int NumSeriesPerChart
		{
			get { return _numSeries; }
			set { SetOnCharts(nameof(NumSeriesPerChart), ref _numSeries, value, c => c.NumSeries = value); }
		}

		private int _numPointsPerChart;
		public int NumPointsPerChart
		{
			get { return _numPointsPerChart; }
			set { SetOnCharts(nameof(NumPointsPerChart), ref _numPointsPerChart, value, c => c.NumPointsPerSeries = value); }
		}



		private int _chartTimeInMs;
		public int ChartTimeInMs
		{
			get { return _chartTimeInMs; }
			set { SetOnCharts(nameof(ChartTimeInMs), ref _chartTimeInMs, value, c => c.ChartTimeInMs = value); }
		}

		private int _timeToSleepInMs;
		public int TimeToSleepInMs
		{
			get { return _timeToSleepInMs; }
			set { SetOnCharts(nameof(TimeToSleepInMs), ref _timeToSleepInMs, value, c => c.TimeToSleepInMs = value); }
		}

		private bool _antiAliasing;
		public bool AntiAliasing
		{
			get { return _antiAliasing; }
			set { SetOnCharts(nameof(AntiAliasing), ref _antiAliasing, value, c => c.AntiAliasing = value); }
		}

		private int _strokeThickness;
		public int StrokeThickness
		{
			get { return _strokeThickness; }
			set { SetOnCharts(nameof(StrokeThickness), ref _strokeThickness, value, c => c.StrokeThickness = value); }
		}

		private bool _useSingleTaskForUpdate;
		public bool UseSingleTaskForUpdate
		{
			get { return _useSingleTaskForUpdate; }
			set { Set(nameof(UseSingleTaskForUpdate), ref _useSingleTaskForUpdate, value); }
		}

		private bool _showDiagnostics;
		public bool ShowDiagnostics
		{
			get { return _showDiagnostics; }
			set { SetOnCharts(nameof(ShowDiagnostics), ref _showDiagnostics, value, c => c.DiagnosticsVisibility = VisFromBool(value)); }
		}

		private ResamplingMode _resamplingMode;
		public ResamplingMode ResamplingMode
		{
			get { return _resamplingMode; }
			set { SetOnCharts(nameof(ResamplingMode), ref _resamplingMode, value, c => c.ResamplingMode = value); }
		}

		public IEnumerable<ResamplingMode> ResamplingModes
		{
			get { return Enum.GetValues(typeof(ResamplingMode)).Cast<ResamplingMode>(); }
		}

		private RenderPriority _renderPriority;
		public RenderPriority RenderPriority
		{
			get { return _renderPriority; }
			set
			{
				bool restartRequired = IsRunning;
				if (restartRequired) { DoStopCharts(); }
				SetOnCharts(nameof(RenderPriority), ref _renderPriority, value, c => c.RenderPriority = value);
				if (restartRequired) { DoStartCharts(); }
			}
		}

		public IEnumerable<RenderPriority> RenderPriorities
		{
			get { return Enum.GetValues(typeof(RenderPriority)).Cast<RenderPriority>(); }
		}

		private double _maxFrameRate = 30.0;
		public double MaxFrameRate
		{
			get { return _maxFrameRate; }
			set { SetOnCharts(nameof(MaxFrameRate), ref _maxFrameRate, value, c => c.MaxFrameRate = value); }
		}

		public IEnumerable<DataFunction> DataFunctions
		{
			get { return Enum.GetValues(typeof(DataFunction)).Cast<DataFunction>(); }
		}
		private DataFunction _dataFunction;
		public DataFunction DataFunction
		{
			get { return _dataFunction; }
			set { SetOnCharts(nameof(DataFunction), ref _dataFunction, value, c => c.DataFunction = value); }
		}

		public ObservableCollection<ChartViewModel> Charts { get; private set; }

		private void InitializeCharts()
		{
			if (_useSingleTaskForUpdate)
			{
				while (NumCharts < Charts.Count)
				{
					Charts.Remove(Charts.Last());
				}

				while (NumCharts > Charts.Count)
				{
					Charts.Add(CreateNewViewModelWithSettings());
				}
			}
			else
			{
				while (NumCharts < Charts.Count)
				{
					var chartToRemove = Charts.Last();
					chartToRemove.Stop();
					Charts.Remove(chartToRemove);
				}

				while (NumCharts > Charts.Count)
				{
					var newChart = CreateNewViewModelWithSettings();
					Charts.Add(newChart);
					if (IsRunning)
					{
						newChart.Start();
					}
				}
			}

			RaisePropertyChanged(nameof(NumRows));
			RaisePropertyChanged(nameof(NumColumns));
		}

		private ChartViewModel CreateNewViewModelWithSettings()
		{
			return new ChartViewModel(NumPointsPerChart)
			{
				AntiAliasing = AntiAliasing,
				ChartTimeInMs = ChartTimeInMs,
				DataFunction = DataFunction,
				DiagnosticsVisibility = VisFromBool(ShowDiagnostics),
				MaxFrameRate = MaxFrameRate,
				NumPointsPerSeries = NumPointsPerChart,
				NumSeries = NumSeriesPerChart,
				RenderPriority = RenderPriority,
				ResamplingMode = ResamplingMode,
				StrokeThickness = StrokeThickness,
				TimeToSleepInMs = TimeToSleepInMs,
			};
		}

		private void SetOnCharts<T>(string propName, ref T field, T value, Action<ChartViewModel> chartSetter)
		{
			if (Set(propName, ref field, value))
			{
				DoOnAllCharts(chartSetter);
			}
		}

		private void DoOnAllCharts(Action<ChartViewModel> action)
		{
			foreach (var c in Charts) { action(c); };
		}

		private void DoStartCharts()
		{
			Log.D("MainVM: Starting charts...");
			IsRunning = true;
			if (_useSingleTaskForUpdate)
			{
				_updateTask = Task.Factory.StartNew(() => UpdateTaskFunction(), TaskCreationOptions.LongRunning);
			}
			else
			{
				DoOnAllCharts(c => c.Start());
			}
		}

		private void DoStopCharts()
		{
			Log.D("MainVM: Stopping charts...");
			IsRunning = false;
			if (_useSingleTaskForUpdate)
			{
				Log.D("Stopping app...");
				if (!IsRunning)
				{
					Log.D("Attempting to stop a non-running app.");
					return;
				}

				// wait at least as long as the time it should sleep before trying to dispose the Task
				if (!_updateTask.Wait(TimeSpan.FromMilliseconds(TimeToSleepInMs * 1.5)))
				{
					try
					{
						_updateTask.Dispose();
					}
					catch (Exception ex)
					{
						Log.D($"Exception on stopping: {ex}");
					}
				}
			}
			else
			{
				DoOnAllCharts(c => c.Stop());
			}
		}

		private static Visibility VisFromBool(bool b)
		{
			return b ? Visibility.Visible : Visibility.Collapsed;
		}

		private bool UpdateTaskFunction()
		{
			while (IsRunning)
			{
				DateTime now = DateTime.Now;
				try
				{
					Dispatcher.CurrentDispatcher.Invoke(
						() => { foreach (var c in Charts) { c.Tick(now); } }
						, DispatcherPriority.Render
					);
					Thread.Sleep(TimeToSleepInMs);
				}
				catch (Exception ex)
				{
					Log.D("Exception: " + ex);
				}
			}

			return true;
		}
	}
}